// 
// Copyright (C) 2019, NinjaTrader LLC <www.ninjatrader.com>.
// NinjaTrader reserves the right to modify or overwrite this NinjaScript component with each release.
//
#region Using declarations
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.DataAnnotations;
using System.Linq;	
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Xml.Serialization;
using NinjaTrader.Gui;
using NinjaTrader.Gui.Chart;
#endregion

namespace NinjaTrader.NinjaScript.DrawingTools
{

	[Gui.CategoryOrder("XABCD Colors", 0)] 
	[Gui.CategoryOrder("XABCD Options", 1)] 
	[Gui.CategoryOrder("Classic Fib Mode", 2)] 
	public abstract class BmdXABCD : DrawingTool
	{
		// this base class is used to draw and manage all three types, so this is used to keep track 
		// of which type of shape is being manipulated. does not need to be localized because not visible on  UI
		protected enum ChartShapeType2
		{
			Unset,
			Triangle2
		}
		
		// for rectangle and ellipse, we show a few fake anchors along edges/corners, so we need to
		// figure out how to actually size the 2 data anchors
		protected enum ResizeMode 
		{
			None,
			TopLeft,
			TopRight,
			BottomLeft,
			BottomRight,
			MoveAll,
		}
		
		private				int							areaOpacity;
		private				Brush						areaBrush;
		private	readonly	DeviceBrush					areaBrushDevice			= new DeviceBrush();
		private	const		double						cursorSensitivity		= 15;
		private				ChartAnchor					editingAnchor;
		private				ChartAnchor 				editingLeftAnchor;
		private				ChartAnchor 				editingTopAnchor; 
		private				ChartAnchor 				editingBottomAnchor;
		private				ChartAnchor 				editingRightAnchor;
		private				ChartAnchor					lastMouseMoveDataPoint	= null;
		private				ResizeMode					resizeMode;
		
		[XmlIgnore] 
		[Display(ResourceType = typeof(Custom.Resource), Name = "NinjaScriptDrawingToolShapesAreaBrush", GroupName = "XABCD Colors", Order = 1)]
		public Brush AreaBrush
		{
			get { return areaBrush; }
			set
			{
				areaBrush = value;
				if (areaBrush != null)
				{
					if (areaBrush.IsFrozen)
						areaBrush = areaBrush.Clone();
					areaBrush.Freeze();
				}
				areaBrushDevice.Brush = null;
			}
		}

		[Browsable(false)]
		public string AreaBrushSerialize
		{
			get { return Serialize.BrushToString(AreaBrush); }
			set { AreaBrush = Serialize.StringToBrush(value); }
		}

		[Range(0,100)]
		[Display(ResourceType = typeof(Custom.Resource), Name = "NinjaScriptDrawingToolAreaOpacity", GroupName = "XABCD Colors", Order = 2)]
		public int AreaOpacity
		{
			get { return areaOpacity; }
			set
			{
				areaOpacity = Math.Max(0, Math.Min(100, value));
				areaBrushDevice.Brush = null;
			}
		} 

		[Display(ResourceType = typeof(Custom.Resource), Name = "NinjaScriptDrawingToolTextOutlineStroke", GroupName = "XABCD Colors", Order = 3)]
		public Stroke				OutlineStroke	{ get; set; }
		
		
		[XmlIgnore()]
		[Display(Name = "Text Color", GroupName = "XABCD Colors", Order = 4)]
		public Brush TextColorBrush
		{ get; set; }
		[Browsable(false)]
        public string TextColorBrushSerializable
        {
            get { return Serialize.BrushToString(TextColorBrush); }
            set { TextColorBrush = Serialize.StringToBrush(value); }
        } 
		
		[XmlIgnore()]
		[Display(Name = "Line Color", GroupName = "XABCD Colors", Order = 4)]
		public Brush LineColorBrush
		{ get; set; }
		[Browsable(false)]
        public string LineColorBrushSerializable
        {
            get { return Serialize.BrushToString(LineColorBrush); }
            set { LineColorBrush = Serialize.StringToBrush(value); }
        }
		[XmlIgnore()]
		[Display(Name = "One to One Color", GroupName = "XABCD Colors", Order = 5)]
		public Brush OneToOneBrush
		{ get; set; }
		[Browsable(false)]
        public string OneToOneBrushSerializable
        {
            get { return Serialize.BrushToString(OneToOneBrush); }
            set { OneToOneBrush = Serialize.StringToBrush(value); }
        }
		[XmlIgnore()]
		[Display(Name = "AOI Color", GroupName = "XABCD Colors", Order = 6)]
		public Brush AOIBrush
		{ get; set; }
		[Browsable(false)]
        public string AOIBrushSerializable
        {
            get { return Serialize.BrushToString(AOIBrush); }
            set { AOIBrush = Serialize.StringToBrush(value); }
        }			
		
//		[NinjaScriptProperty]
//		[Display(ResourceType = typeof(Custom.Resource), Name = "Snap to High/Low", GroupName = "XABCD Options", Order = 1)]
//		public bool SnapToHighLow
//		{ get; set; }
		
		[NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "Show AB One to One", GroupName = "XABCD Options", Order = 2)]
		public bool ShowABOnetoOne
		{ get; set; }				
		
		[Range(0.01, int.MaxValue), NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "Fib Tolerance %", GroupName = "XABCD Options", Order = 3)]
		public double FibTolerance
		{ get; set; }
		
		[NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "Gartley", GroupName = "XABCD Options", Order = 4)]
		public bool showGartley
		{ get; set; }
		[NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "Bat", GroupName = "XABCD Options", Order = 5)]
		public bool showBat
		{ get; set; }
		[NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "Crab", GroupName = "XABCD Options", Order = 6)]
		public bool showCrab
		{ get; set; }
		[NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "Butterfly", GroupName = "XABCD Options", Order = 7)]
		public bool showButterfly
		{ get; set; }		
		
		[NinjaScriptProperty]
		[Display(ResourceType = typeof(Custom.Resource), Name = "(Coming Soon) Classic Fib Mode", GroupName = "Classic Fib Mode", Order = 1)]
		public bool SCRules
		{ get; set; }		

		

		// common to shapes
		[Display(Order = 2)]
		public ChartAnchor			EndAnchor		{ get; set;	}
		[Display(Order = 3)]
		public ChartAnchor			MiddleAnchor	{ get; set; } // only used for triangle
		[Display(Order = 1)]
		public ChartAnchor			StartAnchor		{ get; set; }
		
		[Display(Order = 5)]
		public ChartAnchor			EndAnchor2		{ get; set;	}
		[Display(Order = 6)]
		public ChartAnchor			MiddleAnchor2	{ get; set; } // only used for triangle
	
		

		[Browsable(false)]
		protected ChartShapeType2 	ShapeType 		{ get; set;	}
		
		public override bool SupportsAlerts { get { return true; } }
		
		public override IEnumerable<ChartAnchor> Anchors { get { return ShapeType == ChartShapeType2.Triangle2 ? new[] { StartAnchor, MiddleAnchor, EndAnchor,  MiddleAnchor2, EndAnchor2 } : new[] { StartAnchor, EndAnchor2 }; } }

		public override void OnCalculateMinMax()
		{
			MinValue = double.MaxValue;
			MaxValue = double.MinValue;

			if (!IsVisible)
				return;

			// return min/max values only if something has been actually drawn
			if (Anchors.Any(a => !a.IsEditing))
				foreach (ChartAnchor anchor in Anchors)
				{
					MinValue = Math.Min(anchor.Price, MinValue);
					MaxValue = Math.Max(anchor.Price, MaxValue);
				}
		}
		
		private SharpDX.Direct2D1.PathGeometry CreateTriangleGeometry(ChartControl chartControl, ChartPanel chartPanel, ChartScale chartScale, double pixelAdjust)
		{
			Point startPoint			= StartAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point midPoint				= MiddleAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point endPoint 				= EndAnchor.GetPoint(chartControl, chartPanel, chartScale);
			
			Point midPoint2				= MiddleAnchor2.GetPoint(chartControl, chartPanel, chartScale);
			Point endPoint2 			= EndAnchor2.GetPoint(chartControl, chartPanel, chartScale);
			
			Vector pixelAdjustVec		= new Vector(pixelAdjust, pixelAdjust);
			SharpDX.Vector2 startVec 	= (startPoint + pixelAdjustVec).ToVector2();
			SharpDX.Vector2 midVec 		= (midPoint + pixelAdjustVec).ToVector2();
			SharpDX.Vector2 endVec 		= (endPoint + pixelAdjustVec).ToVector2();
			
			SharpDX.Vector2 midVec2 	= (midPoint2 + pixelAdjustVec).ToVector2();
			SharpDX.Vector2 endVec2 	= (endPoint2 + pixelAdjustVec).ToVector2();			
			
			SharpDX.Direct2D1.PathGeometry pathGeometry = new SharpDX.Direct2D1.PathGeometry(Core.Globals.D2DFactory);
			SharpDX.Direct2D1.GeometrySink geometrySink = pathGeometry.Open();
			geometrySink.BeginFigure(startVec, SharpDX.Direct2D1.FigureBegin.Filled);

			geometrySink.AddLines(new[] 
			{
				startVec, midVec, 	// start -> mid,
				midVec, endVec,		// mid -> top
				endVec, startVec, 	// top -> start (cap it off)
				endVec, midVec2,
				midVec2, endVec2,		// mid -> top
				endVec2, endVec				
				
			});
				
			geometrySink.EndFigure(SharpDX.Direct2D1.FigureEnd.Open);
			geometrySink.Close(); // calls dispose for you
			return pathGeometry;
		}

		protected override void Dispose(bool disposing)
		{
			base.Dispose(disposing);
			if (areaBrushDevice != null)
				areaBrushDevice.RenderTarget = null;
		}

		

		public override Cursor GetCursor(ChartControl chartControl, ChartPanel chartPanel, ChartScale chartScale, Point point)
		{
			
			if (DrawingState == DrawingState.Building)
				return Cursors.Pen;
			
			if (DrawingState == DrawingState.Moving)
				return IsLocked ? Cursors.No : Cursors.SizeAll;
			
			if (DrawingState == DrawingState.Editing && IsLocked)
				return Cursors.No;
			
			// work like NT7 here, if we're on the edge of a shape, or dead center, move

			if (ShapeType == ChartShapeType2.Triangle2)
			{
				// triangle shows actual anchor points so much easier
				ChartAnchor closest = GetClosestAnchor(chartControl, chartPanel, chartScale, cursorSensitivity, point);
				if (closest == null)
				{
					// get the centroid point, see if we're near that
					Point[] trianglePoints = GetTriangleAnchorPoints(chartControl,chartScale, true);
					if ((trianglePoints.Last() - point).Length <= cursorSensitivity)
						return IsLocked ? Cursors.Arrow : Cursors.SizeAll;
					// check points as vectors to each other to check each edge (dont use center point which is last in the array)
					for (int i = 0; i < 3; ++i)
					{
						Point startPoint = trianglePoints[i == 2 ? 0 : i + 1];
						Vector vec = trianglePoints[i] - startPoint;
						if (MathHelper.IsPointAlongVector(point, startPoint, vec, 10))
							return IsLocked ? Cursors.Arrow : Cursors.SizeAll;
					}

					return null;
				}
				return 	IsLocked ? null : Cursors.SizeNESW;
			}
			
			bool isRect = false;
			
			ResizeMode tmpResizeMode = resizeMode != ResizeMode.None ? resizeMode : GetResizeModeForPoint(point, chartControl, chartScale, DrawingState == DrawingState.Normal);
			switch (tmpResizeMode)
			{
				// ellipse rect is shifted to centerpoints of each rectangel edge,
				// so we wanna use up/down/left/right variants instead for it
				case ResizeMode.TopLeft: 		// fall through
				case ResizeMode.BottomRight:	return	IsLocked ? Cursors.Arrow : isRect ? Cursors.SizeNWSE : Cursors.SizeNS;
				case ResizeMode.TopRight: 		// fall through
				case ResizeMode.BottomLeft: 	return	IsLocked ? Cursors.Arrow : isRect ? Cursors.SizeNESW : Cursors.SizeWE;
				case ResizeMode.MoveAll:		return	IsLocked ? Cursors.Arrow : Cursors.SizeAll;
			}
			return null;
		}
		
		private static Point? GetClosestPoint(IEnumerable<Point> inputPoints, Point desired, bool useSensitivity)
		{
			IOrderedEnumerable<Point> ordered = inputPoints.OrderBy(pt => (pt - desired).Length);
			Point closestPoint = ordered.First();
			if (useSensitivity && (closestPoint - desired).Length > cursorSensitivity)
				return null;
			return closestPoint;
		}
		
		
		private ResizeMode GetResizeModeForPoint(Point pt, ChartControl chartControl, ChartScale chartScale, bool useCursorSens)
		{
			switch (ShapeType)
			{
			}
			return ResizeMode.None;
		}
	
		public sealed override Point[] GetSelectionPoints(ChartControl chartControl, ChartScale chartScale)
		{
			switch (ShapeType)
			{
				
				case ChartShapeType2.Triangle2:
					return GetTriangleAnchorPoints(chartControl, chartScale, true);
			}
			return new Point[0];
		}
		
		private Point[] GetTriangleAnchorPoints(ChartControl chartControl, ChartScale chartScale, bool includeCentroid)
		{
			ChartPanel	chartPanel	= chartControl.ChartPanels[PanelIndex];
			Point		midPoint 	= MiddleAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point		startPoint	= StartAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point		endPoint	= EndAnchor.GetPoint(chartControl, chartPanel, chartScale);
			
			Point		midPoint2 	= MiddleAnchor2.GetPoint(chartControl, chartPanel, chartScale);
			Point		endPoint2	= EndAnchor2.GetPoint(chartControl, chartPanel, chartScale);

			return includeCentroid
				? new[] { startPoint, midPoint, endPoint, new Point((startPoint.X + midPoint.X + endPoint.X)/3d, (startPoint.Y + midPoint.Y + endPoint.Y)/3d),  midPoint2, endPoint2 }
				: new[] { startPoint, midPoint, endPoint,  midPoint2, endPoint2 };
		}
	
		public override IEnumerable<AlertConditionItem> GetAlertConditionItems()
		{
			// nothing fancy here, but we do need to return at least 1 condition item to check an
			// alert on to be considered valid
			yield return new AlertConditionItem 
			{
				Name					= "Shape area",
				ShouldOnlyDisplayName	= true,
			};
		}
		
		public override IEnumerable<Condition> GetValidAlertConditions()
		{
			return new[] { Condition.CrossInside, Condition.CrossOutside };
		}
		
		public override bool IsAlertConditionTrue(AlertConditionItem conditionItem, Condition condition, ChartAlertValue[] values, ChartControl chartControl, ChartScale chartScale)
		{
			// first make sure the values fall in our chart object at all
			double		minPrice	= Anchors.Min(a => a.Price);
			double		maxPrice	= Anchors.Max(a => a.Price);
			DateTime	minTime		= Anchors.Min(a => a.Time);
			DateTime	maxTime		= Anchors.Max(a => a.Time);

			ChartPanel	chartPanel	= chartControl.ChartPanels[PanelIndex];
			Point		startPoint	= StartAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point		endPoint	= EndAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point		centerPoint	= startPoint + ((endPoint - startPoint) * 0.5);
				
			Predicate<ChartAlertValue>	predicate;
			Func<ChartAlertValue,Point> getBarPoint = v => v.ValueType != ChartAlertValueType.StaticValue ? 
				new Point(chartControl.GetXByTime(v.Time),chartScale.GetYByValue(v.Value)) : new Point(0,0); 

			switch (ShapeType)
			{
				
				case ChartShapeType2.Triangle2:
					Point[] trianglePoints = GetTriangleAnchorPoints(chartControl, chartScale, false);
					predicate = v =>
					{
						bool isInside = MathHelper.IsPointInsideTriangle(getBarPoint(v), trianglePoints[0], trianglePoints[1], trianglePoints[2]);
						return condition == Condition.CrossInside ? isInside : !isInside;
					};
					break;
				default: return false;
			}
		
			return MathHelper.DidPredicateCross(values, predicate);
		}
		
		public override bool IsVisibleOnChart(ChartControl chartControl, ChartScale chartScale, DateTime firstTimeOnChart, DateTime lastTimeOnChart)
		{
			if (DrawingState == DrawingState.Building)
				return true;

			float minX = float.MaxValue;
			float maxX = float.MinValue;
			ChartPanel chartPanel = chartControl.ChartPanels[PanelIndex];
			// create an axis aligned bounding box taking into acct all 3 points for triangle to use for visibility checks
			foreach (Point pt in Anchors.Select(a => a.GetPoint(chartControl, chartPanel, chartScale)))
			{
				minX = (float)Math.Min(minX, pt.X);
				maxX = (float)Math.Max(maxX, pt.X);
			}

			DateTime	leftWidthTime	= chartControl.GetTimeByX((int) minX);
			DateTime	rightWidthTime	= chartControl.GetTimeByX((int) maxX);
			
			// check our width is visible somewhere horizontally
			return leftWidthTime <= lastTimeOnChart && rightWidthTime >= firstTimeOnChart;
		}

		public override void OnMouseDown(ChartControl chartControl, ChartPanel chartPanel, ChartScale chartScale, ChartAnchor dataPoint)
		{
			
			if (ShapeType == ChartShapeType2.Unset){
				return;
			}

			switch (DrawingState)
			{
				case DrawingState.Building:
					if (StartAnchor.IsEditing)
					{							
						//StartAnchor.Price
						// give mid & end anchor something to start with so we dont try to render it with bad values right away
						dataPoint.CopyDataValues(StartAnchor);
						dataPoint.CopyDataValues(MiddleAnchor);
						dataPoint.CopyDataValues(EndAnchor);
						dataPoint.CopyDataValues(MiddleAnchor2);
						dataPoint.CopyDataValues(EndAnchor2);
						StartAnchor.IsEditing = false;
					}
					else if (ShapeType == ChartShapeType2.Triangle2 && MiddleAnchor.IsEditing)
					{
						dataPoint.CopyDataValues(MiddleAnchor);
						MiddleAnchor.IsEditing = false;
					}
					else if (EndAnchor.IsEditing)
					{
						dataPoint.CopyDataValues(EndAnchor);
						EndAnchor.IsEditing = false;
					}
					else if (MiddleAnchor2.IsEditing)
					{
						dataPoint.CopyDataValues(MiddleAnchor2);
						MiddleAnchor2.IsEditing = false;
					}
					else if (EndAnchor2.IsEditing)
					{
						dataPoint.CopyDataValues(EndAnchor2);
						EndAnchor2.IsEditing = false;
					}
					
					// is initial building done? (all anchors set)
					if (!EndAnchor2.IsEditing)
					{
						// if we're a triangle, check middle is done too
						if (!MiddleAnchor2.IsEditing)
						{									
							DrawingState 	= DrawingState.Normal;
							IsSelected 		= false; 
						}
					}

					break;
				case DrawingState.Normal:
					Point point = dataPoint.GetPoint(chartControl, chartPanel, chartScale);
					switch (ShapeType)
					{
						case ChartShapeType2.Triangle2:
							editingAnchor = GetClosestAnchor(chartControl, chartPanel, chartScale, cursorSensitivity, point);
							
							if (editingAnchor != null)
							{
								editingAnchor.IsEditing = true;
								DrawingState = DrawingState.Editing;
							}
							else
							{
								if (GetCursor(chartControl, chartPanel, chartScale, point) != null)
									DrawingState = DrawingState.Moving;
								else
								{
									Point[] trianglePoints = GetTriangleAnchorPoints(chartControl,chartScale, true);
									if (!MathHelper.IsPointInsideTriangle(point, trianglePoints[0], trianglePoints[1], trianglePoints[2]))
										IsSelected = false;
								}
							}
							break;
						
					}
					if (lastMouseMoveDataPoint == null)
						lastMouseMoveDataPoint = new ChartAnchor();
					dataPoint.CopyDataValues(lastMouseMoveDataPoint);
					break;
			}
		}
		
		public override void OnMouseMove(ChartControl chartControl, ChartPanel chartPanel, ChartScale chartScale, ChartAnchor dataPoint)
		{
			
			if (ShapeType == ChartShapeType2.Unset || IsLocked && DrawingState != DrawingState.Building)
				return;
			
			if (DrawingState == DrawingState.Building)
			{
				// just simply update both these anchors as user moves mouse, end being same as middle is ok for triangle
				if (MiddleAnchor.IsEditing)
					dataPoint.CopyDataValues(MiddleAnchor);
				if (EndAnchor.IsEditing)
					dataPoint.CopyDataValues(EndAnchor);
				if (MiddleAnchor2.IsEditing)
					dataPoint.CopyDataValues(MiddleAnchor2);
				if (EndAnchor2.IsEditing)
					dataPoint.CopyDataValues(EndAnchor2);
			}
			else if (DrawingState == DrawingState.Editing)
			{
				// triangle mode: editing actual anchor, easy enough
				if (ShapeType == ChartShapeType2.Triangle2 && editingAnchor != null)
					dataPoint.CopyDataValues(editingAnchor);
				else
				{
					if (lastMouseMoveDataPoint == null)
						lastMouseMoveDataPoint = new ChartAnchor();
					switch (resizeMode)
					{
						case ResizeMode.TopLeft:
							editingTopAnchor.Price = lastMouseMoveDataPoint.Price;
							// only update Y on ellipse because this is actually the top edge anchor
							if (ShapeType == ChartShapeType2.Triangle2)
							{
								editingLeftAnchor.SlotIndex = lastMouseMoveDataPoint.SlotIndex;
								editingLeftAnchor.Time = lastMouseMoveDataPoint.Time;

								dataPoint.CopyDataValues(lastMouseMoveDataPoint);
							}							
							break;
						case ResizeMode.BottomRight: 
							// only update Y on ellipse because this is actually the bottom edge anchor
							editingBottomAnchor.Price = lastMouseMoveDataPoint.Price;
							if (ShapeType == ChartShapeType2.Triangle2)
							{
								editingRightAnchor.Time = lastMouseMoveDataPoint.Time;
								editingRightAnchor.SlotIndex = lastMouseMoveDataPoint.SlotIndex;
								dataPoint.CopyDataValues(lastMouseMoveDataPoint);
							}
							
							break;
						case ResizeMode.TopRight:
							// on ellipse top right is actually right edge, so dont move top up/down
							editingRightAnchor.SlotIndex = lastMouseMoveDataPoint.SlotIndex;
							editingRightAnchor.Time = lastMouseMoveDataPoint.Time;
							if (ShapeType == ChartShapeType2.Triangle2)
							{
								editingTopAnchor.Price = lastMouseMoveDataPoint.Price;
								dataPoint.CopyDataValues(lastMouseMoveDataPoint);
							}
							
							break;
						case ResizeMode.BottomLeft:
							// no real anchor to simply move, figure out which is currently 'left' anchor
							editingLeftAnchor.Time = lastMouseMoveDataPoint.Time;
							editingLeftAnchor.SlotIndex = lastMouseMoveDataPoint.SlotIndex;
							if (ShapeType == ChartShapeType2.Triangle2)
							{
								editingBottomAnchor.Price = lastMouseMoveDataPoint.Price;
								dataPoint.CopyDataValues(lastMouseMoveDataPoint);
							}
							
							break;
					}
				}
			}
			else if (DrawingState == DrawingState.Moving)
			{
				foreach (ChartAnchor anchor in Anchors)
					anchor.MoveAnchor(InitialMouseDownAnchor, dataPoint, chartControl, chartPanel, chartScale, this);
			}
		}
		
		public override void OnMouseUp(ChartControl chartControl, ChartPanel chartPanel, ChartScale chartScale, ChartAnchor dataPoint)
		{
			if (DrawingState == DrawingState.Building) 
				return;

			lastMouseMoveDataPoint	= null;
			DrawingState			= DrawingState.Normal;
			editingAnchor			= null;
			editingLeftAnchor		= null;
			editingTopAnchor		= null;
			editingRightAnchor		= null;
			editingBottomAnchor		= null;
			resizeMode				= ResizeMode.None;
		}
		
		public override void OnRender(ChartControl chartControl, ChartScale chartScale)
		{
			if (ShapeType == ChartShapeType2.Unset)
				return;
			Stroke outlineStroke		= OutlineStroke;
			RenderTarget.AntialiasMode	= SharpDX.Direct2D1.AntialiasMode.PerPrimitive;
			outlineStroke.RenderTarget	= RenderTarget;
			ChartPanel chartPanel		= chartControl.ChartPanels[PanelIndex];
			Point startPoint			= StartAnchor.GetPoint(chartControl, chartPanel, chartScale);
			Point endPoint				= EndAnchor.GetPoint(chartControl, chartPanel, chartScale);

			double width				= endPoint.X - startPoint.X;
			double height				= endPoint.Y - startPoint.Y;
			
			SharpDX.Vector2 centerPoint	= (startPoint + ((endPoint - startPoint) / 2)).ToVector2();
			
			// dont bother with an area brush if we're doing a hit test (software) render pass. we do not render area then.
			// this allows us to select something behind our area brush (like NT7)
			if (!IsInHitTest && AreaBrush != null)
			{
				if (areaBrushDevice.Brush == null)
				{
					Brush brushCopy			= areaBrush.Clone();
					brushCopy.Opacity		= areaOpacity / 100d; 
					areaBrushDevice.Brush	= brushCopy;
				}
				areaBrushDevice.RenderTarget = RenderTarget;
			}
			else
			{
				areaBrushDevice.RenderTarget = null;
				areaBrushDevice.Brush = null;
			}
			
			// align to full pixel to avoid unneeded aliasing
			double strokePixAdjust =	outlineStroke.Width % 2 == 0 ? 0.5d : 0d;
			
			switch (ShapeType)
			{				
				case ChartShapeType2.Triangle2:
				{
					// always re-create triangle geo here
					SharpDX.Direct2D1.PathGeometry triGeo = CreateTriangleGeometry(chartControl, chartPanel, chartScale, strokePixAdjust);
					if (!IsInHitTest && areaBrushDevice.BrushDX != null)
						RenderTarget.FillGeometry(triGeo, areaBrushDevice.BrushDX);
					else 
					{
						// Triangle can be selected by center anchor point still, so give something for the hit test pass to 
						// detect we want to be hit test there, so draw a rect in center. actual brush doesnt matter
						Point centroid = GetTriangleAnchorPoints(chartControl, chartScale, true).Last();
						RenderTarget.FillRectangle(new SharpDX.RectangleF((float)centroid.X - 5f, (float)centroid.Y - 5f, (float)cursorSensitivity, (float)cursorSensitivity), chartControl.SelectionBrush);
					}
					SharpDX.Direct2D1.Brush tmpBrush = IsInHitTest ? chartControl.SelectionBrush : outlineStroke.BrushDX;
					RenderTarget.DrawGeometry(triGeo, tmpBrush, outlineStroke.Width, outlineStroke.StrokeStyle);
					triGeo.Dispose();
					break;
				}
			}
			
			SCRules = false;
			
			ChartBars bars = GetAttachedToChartBars();
			SharpDX.DirectWrite.TextFormat textFormat = chartControl.Properties.LabelFont.ToDirectWriteTextFormat();		
			
			double xCoordX = chartControl.GetXByBarIndex(bars, (int)StartAnchor.SlotIndex);
			double yCoordX = chartScale.GetYByValue(StartAnchor.Price);			
			double xCoordA = chartControl.GetXByBarIndex(bars, (int)MiddleAnchor.SlotIndex);
			double yCoordA = chartScale.GetYByValue(MiddleAnchor.Price);
			double xCoordB = chartControl.GetXByBarIndex(bars, (int)EndAnchor.SlotIndex);
			double yCoordB = chartScale.GetYByValue(EndAnchor.Price);			
			double xCoordC = chartControl.GetXByBarIndex(bars, (int)MiddleAnchor2.SlotIndex);
			double yCoordC = chartScale.GetYByValue(MiddleAnchor2.Price);
			double xCoordD = chartControl.GetXByBarIndex(bars, (int)EndAnchor2.SlotIndex);
			double yCoordD = chartScale.GetYByValue(EndAnchor2.Price);	
					
			double priceX = StartAnchor.Price;
			double priceA = MiddleAnchor.Price;
			double priceB = EndAnchor.Price;
			double priceC = MiddleAnchor2.Price;
			double priceD = EndAnchor2.Price;

			bool bullish = (MiddleAnchor.Price - StartAnchor.Price ) > 0  ? true : false;
			
//			if(SnapToHighLow){				
//				priceX = yCoordA < yCoordX ? bars.Bars.GetLow((int)StartAnchor.SlotIndex) : bars.Bars.GetHigh((int)StartAnchor.SlotIndex);				
//				priceA = yCoordA > yCoordX ? bars.Bars.GetLow((int)MiddleAnchor.SlotIndex) : bars.Bars.GetHigh((int)MiddleAnchor.SlotIndex);
//				priceB = yCoordA < yCoordX ? bars.Bars.GetLow((int)EndAnchor.SlotIndex) : bars.Bars.GetHigh((int)EndAnchor.SlotIndex);
//				priceC = yCoordA > yCoordX ? bars.Bars.GetLow((int)MiddleAnchor2.SlotIndex) : bars.Bars.GetHigh((int)MiddleAnchor2.SlotIndex);
//				priceD = yCoordA < yCoordX ? bars.Bars.GetLow((int)EndAnchor2.SlotIndex) : bars.Bars.GetHigh((int)EndAnchor2.SlotIndex);
				
//				yCoordX = chartScale.GetYByValue(priceX);
//				yCoordA = chartScale.GetYByValue(priceA);
//				yCoordB = chartScale.GetYByValue(priceB);
//				yCoordC = chartScale.GetYByValue(priceC);
//			}		
			
			
			int offset = bullish ? 15 : -15;						
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) {					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordX-3, (float)yCoordX + offset, 200, 200);
				String text = "X"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordA-3, (float)yCoordA + (offset*-1), 200, 200);
				String text = "A"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordB-3, (float)yCoordB + offset, 200, 200);
				String text = "B"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordC-3, (float)yCoordC + (offset*-1), 200, 200);
				String text = "C"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordD, (float)yCoordD + offset, 200, 200);
				String text = "D"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}			
			
			//X to A points
			double swingLength = bullish ? priceA - priceX  :  priceX - priceA;	
			offset = bullish ? offset + 10 : offset-10;
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordX-10, (float)yCoordX + offset, 200, 200);
				String text =  Math.Round(swingLength, 0) +" pts"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			
			
			double retraceXA1 = 0;
			double retraceXA2 = 0;
			double retraceAB = 0;

			
			//X to A Retracement (Drawn on X to D)
			swingLength = bullish ? priceA - priceX  :  priceX - priceA;			
			retraceXA1 = bullish ? (priceA - EndAnchor2.Price) / swingLength : (EndAnchor2.Price - priceA) / swingLength;			
			double xToDMidPointX = (xCoordX + xCoordD) / 2;
			double xToDmidPointY = (yCoordX + yCoordD) / 2;
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xToDMidPointX, (float)xToDmidPointY, 200, 200);
				String text = Math.Round(retraceXA1, 2)+"%"; 				
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			Point p1 = new Point(xCoordX, yCoordX);
			Point p2 = new Point(xCoordD, yCoordD);
			using (SharpDX.Direct2D1.Brush dxBrush = LineColorBrush.ToDxBrush(RenderTarget)) { 	
				RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush);
			}
			
			//X to A Retracement (drawn on X to B)
			swingLength = bullish ? priceA - priceX  :  priceX - priceA;			
			retraceXA2 = bullish ? (priceA - priceB) / swingLength : (priceB - priceA) / swingLength;		
			double xToBMidPointX = (xCoordB + xCoordX) / 2;
			double xToBmidPointY = (yCoordB + yCoordX) / 2;
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xToBMidPointX, (float)xToBmidPointY, 200, 200);
				String text = Math.Round(retraceXA2, 3)+"%"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			
			//A to B Retracement
			swingLength = bullish ? priceA - priceB  :  priceB - priceA;			
			retraceAB = bullish ? (priceC - priceB) / swingLength : (priceB - priceC) / swingLength;		
			double aToCMidPointX = (xCoordA + xCoordC) / 2;
			double aToCmidPointY = (yCoordA + yCoordC) / 2;
			using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
				SharpDX.RectangleF rect = new SharpDX.RectangleF((float)aToCMidPointX, (float)aToCmidPointY, 200, 200);
				String text =  Math.Round(retraceAB, 3)+"%"; 							
				RenderTarget.DrawText(text, textFormat, rect, dxBrush);
			}
			p1 = new Point(xCoordA, yCoordA);
			p2 = new Point(xCoordC, yCoordC);
			using (SharpDX.Direct2D1.Brush dxBrush = LineColorBrush.ToDxBrush(RenderTarget)) { 	
				RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush);
			}						

		
			//A to B Projection (labelled on B to D)
			double retracement = 0;
			if(!SCRules){
				swingLength = bullish ? priceA - priceB  :  priceB - priceA;	
				retracement = bullish ? (priceA - EndAnchor2.Price) / swingLength : (EndAnchor2.Price - priceA) / swingLength;
				retracement =  Math.Round(retracement, 2);
				double bToDMidPointX = (xCoordB + xCoordD) / 2;
				double bToDmidPointY = (yCoordB + yCoordD) / 2;
				using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
					SharpDX.RectangleF rect = new SharpDX.RectangleF((float)bToDMidPointX, (float)bToDmidPointY, 200, 200);
					String text = retracement+"%"; 							
					RenderTarget.DrawText(text, textFormat, rect, dxBrush);
				}
			} 
			//C to B Projection (labelled on B to D)
			else if(SCRules){
				swingLength = bullish ? priceC - priceB  :  priceB - priceC;	
				retracement = bullish ? (priceC - EndAnchor2.Price) / swingLength : (EndAnchor2.Price - priceC) / swingLength;
				retracement =  Math.Round(retracement, 2);
				double bToDMidPointX = (xCoordB + xCoordD) / 2;
				double bToDmidPointY = (yCoordB + yCoordD) / 2;
				using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
					SharpDX.RectangleF rect = new SharpDX.RectangleF((float)bToDMidPointX, (float)bToDmidPointY, 200, 200);
					String text = retracement+"%"; 							
					RenderTarget.DrawText(text, textFormat, rect, dxBrush);
				}
			}
			
			// one to one projection of AB
			double oneToOneXAmt = 0;
			double oneToOneCompletePrice = 0;		
			double oneToOnePriceChange = priceB - priceA;
			oneToOneCompletePrice = priceC + oneToOnePriceChange;
			oneToOneXAmt =  chartControl.GetXByBarIndex(bars, (int)EndAnchor.SlotIndex) -  chartControl.GetXByBarIndex(bars, (int)MiddleAnchor.SlotIndex);
			if(ShowABOnetoOne){
				p1 = new Point(xCoordC, chartScale.GetYByValue(priceC));
				p2 = new Point(xCoordC + oneToOneXAmt, chartScale.GetYByValue(oneToOneCompletePrice));
				using (SharpDX.Direct2D1.Brush dxBrush = OneToOneBrush.ToDxBrush(RenderTarget)) { 	
					RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush);
				}
			}

			
			bool foundGartley = false;
			bool foundBat = false;
			bool foundCrab = false;
			bool foundButterfly = false;
			
			//Gartley
			if(showGartley){
			if( (!SCRules && retraceXA2 >= .618-FibTolerance && retraceXA2 <= .786+FibTolerance)  || (SCRules && retraceXA2 >= (.618 - FibTolerance)  && retraceXA2 <= (.618 + FibTolerance))  ){
				
				if( (!SCRules && retraceAB >= .618-FibTolerance && retraceAB <= .786+FibTolerance) || (SCRules && retraceAB >= .382-FibTolerance && retraceAB <= .886+FibTolerance)  ){
					foundGartley = true;
								
					double aoi1=0;
					double aoi2=0;
					double aoi3=0;					
					
					swingLength = bullish ? priceA - priceX  :  priceX - priceA;
					aoi1 = bullish ? priceA - (swingLength * .786) : priceA + (swingLength * .786); 					

					if(!SCRules){
						swingLength = bullish ? priceA - priceB  :  priceB - priceA;	
						aoi2 = bullish ? priceA - (swingLength * 1.272) : priceA + (swingLength * 1.272);	
						aoi3 = aoi2;
					} else if(SCRules){
						swingLength = bullish ? priceC - priceB  :  priceB - priceC;
						aoi2 = bullish ? priceC - (swingLength * 1.13) : priceC + (swingLength * 1.13);
						aoi3 = bullish ? priceC - (swingLength * 1.618) : priceC + (swingLength * 1.618);	
					}
					
					double highest = aoi1 > aoi2 ? aoi1 : aoi2;
					highest = aoi3 > highest ? aoi3 : highest;
					highest = oneToOneCompletePrice > highest ? oneToOneCompletePrice : highest;
					highest = !bullish && highest > priceX ? priceX : highest;					

					double lowest = aoi1 < aoi2 ? aoi1 : aoi2;
					lowest = aoi3 < lowest ? aoi3 : lowest;	
					lowest = oneToOneCompletePrice < lowest ? oneToOneCompletePrice : lowest;
					lowest = bullish && lowest < priceX ? priceX : lowest;

					SharpDX.Direct2D1.StrokeStyle strokeStyle = new Stroke(AOIBrush, DashStyleHelper.Dash, 1).StrokeStyle;
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(highest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(highest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 1.3f, strokeStyle);
					}
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(lowest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(lowest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 1.3f, strokeStyle);
					}
				}
			}
			}
			
			//BAT
			if(showBat){
			if(retraceXA2 >= .382-FibTolerance && retraceXA2 <= .500+FibTolerance){				
				if( (!SCRules && retraceAB >= .618-FibTolerance && retraceAB <= .886+FibTolerance) || (SCRules && retraceAB >= .382-FibTolerance && retraceAB <= .886+FibTolerance)  ){
					foundBat = true;
								
					double aoi1=0;
					double aoi2=0;
					double aoi3=0;					
					
					swingLength = bullish ? priceA - priceX  :  priceX - priceA;
					aoi1 = bullish ? priceA - (swingLength * .886) : priceA + (swingLength * .886); 					

					if(!SCRules){
						swingLength = bullish ? priceA - priceB  :  priceB - priceA;	
						aoi2 = bullish ? priceA - (swingLength * 1.618) : priceA + (swingLength * 1.618);	
						aoi3 = aoi2;
					} else if(SCRules){
						swingLength = bullish ? priceC - priceB  :  priceB - priceC;
						aoi2 = bullish ? priceC - (swingLength * 2.618) : priceC + (swingLength * 2.618);
						aoi3 = bullish ? priceC - (swingLength * 1.618) : priceC + (swingLength * 1.618);
						aoi2 = bullish ? (aoi2 < aoi1 ? aoi1 : aoi2) : (aoi2 > aoi1 ? aoi1 : aoi2); //i.e. 2.618 should not be more than .886 of XA
					}
					
					double highest = aoi1 > aoi2 ? aoi1 : aoi2;
					highest = aoi3 > highest ? aoi3 : highest;
					double lowest = aoi1 < aoi2 ? aoi1 : aoi2;
					lowest = aoi3 < lowest ? aoi3 : lowest;	

					SharpDX.Direct2D1.StrokeStyle strokeStyle = new Stroke(AOIBrush, DashStyleHelper.Dash, 1).StrokeStyle;
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(highest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(highest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 1.3f, strokeStyle);
					}
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(lowest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(lowest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 1.3f, strokeStyle);
					}
				}
			}
			}
			
			//Crab
			if(showCrab){
			if( (!SCRules && retraceXA2 >= .500-FibTolerance && retraceXA2 <= .618+FibTolerance)  || (SCRules && retraceXA2 >= .382-FibTolerance && retraceXA2 <= .618+FibTolerance)  ){			
				if( (!SCRules && retraceAB >= .500-FibTolerance && retraceAB <= .618+FibTolerance) || (SCRules && retraceAB >= .382-FibTolerance && retraceAB <= .886+FibTolerance)  ){
					foundCrab = true;
								
					double aoi1=0;
					double aoi2=0;
					double aoi3=0;					
					
					swingLength = bullish ? priceA - priceX  :  priceX - priceA;
					aoi1 = bullish ? priceA - (swingLength * 1.618) : priceA + (swingLength * 1.618); 					

					if(!SCRules){
						swingLength = bullish ? priceA - priceB  :  priceB - priceA;	
						aoi2 = bullish ? priceA - (swingLength * 2.618) : priceA + (swingLength * 2.618);	
						aoi3 = aoi2;
					} else if(SCRules){
						swingLength = bullish ? priceC - priceB  :  priceB - priceC;
						aoi2 = bullish ? priceC - (swingLength * 2.618) : priceC + (swingLength * 2.618);
						aoi3 = bullish ? priceC - (swingLength * 3.618) : priceC + (swingLength * 3.618);	
					}
					
					double highest = aoi1 > aoi2 ? aoi1 : aoi2;
					highest = aoi3 > highest ? aoi3 : highest;
					double lowest = aoi1 < aoi2 ? aoi1 : aoi2;
					lowest = aoi3 < lowest ? aoi3 : lowest;	

					SharpDX.Direct2D1.StrokeStyle strokeStyle = new Stroke(AOIBrush, DashStyleHelper.DashDotDot, 1).StrokeStyle;
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(highest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(highest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 2.3f, strokeStyle);
					}
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(lowest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(lowest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 2.3f, strokeStyle);
					}
				}
			}
			}
			
			//Butterfly
			if(showButterfly){
			if( retraceXA2 >= (.786 - FibTolerance)  && retraceXA2 <= (.786 + FibTolerance) ){			
				if( (!SCRules && retraceAB >= .500-FibTolerance && retraceAB <= .786+FibTolerance) || (SCRules && retraceAB >= .382-FibTolerance && retraceAB <= .886+FibTolerance)  ){
					foundButterfly = true;
								
					double aoi1=0;
					double aoi2=0;
					double aoi3=0;					
					
					swingLength = bullish ? priceA - priceX  :  priceX - priceA;
					aoi1 = bullish ? priceA - (swingLength * 1.272) : priceA + (swingLength * 1.272); 					

					if(!SCRules){
						swingLength = bullish ? priceA - priceB  :  priceB - priceA;	
						aoi2 = bullish ? priceA - (swingLength * 1.618) : priceA + (swingLength * 1.618);	
						aoi3 = aoi2;
					} else if(SCRules){
						swingLength = bullish ? priceC - priceB  :  priceB - priceC;
						aoi2 = bullish ? priceC - (swingLength * 1.618) : priceC + (swingLength * 1.618);
						aoi3 = bullish ? priceC - (swingLength * 2.24) : priceC + (swingLength * 2.24);	
					}
					
					double highest = aoi1 > aoi2 ? aoi1 : aoi2;
					highest = aoi3 > highest ? aoi3 : highest;
					double lowest = aoi1 < aoi2 ? aoi1 : aoi2;
					lowest = aoi3 < lowest ? aoi3 : lowest;	

					SharpDX.Direct2D1.StrokeStyle strokeStyle = new Stroke(AOIBrush, DashStyleHelper.DashDotDot, 1).StrokeStyle;
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(highest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(highest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 2.3f, strokeStyle);
					}
					
					p1 = new Point(xCoordC, chartScale.GetYByValue(lowest));
					p2 = new Point(xCoordC + (oneToOneXAmt * 6), chartScale.GetYByValue(lowest));
					using (SharpDX.Direct2D1.Brush dxBrush = AOIBrush.ToDxBrush(RenderTarget)) { 	
						RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush, 2.3f, strokeStyle);
					}
				}
			}
			}
			
			
			if((foundGartley && showGartley) || (foundBat && showBat) || (foundCrab && showCrab) || (foundButterfly && showButterfly)){
				//A to D Retrace BE/Target
				swingLength = bullish ? priceA - priceD  :  priceD - priceA;			
				double three82 = bullish ? (priceD + (swingLength * .382)) : (priceD - (swingLength * .382));		
				double six18 = bullish ? (priceD + (swingLength * .618)) : (priceD - (swingLength * .618));

				using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
					SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordD, chartScale.GetYByValue(three82), 200, 200);
					String text =  ".382 (BE)"; 							
					RenderTarget.DrawText(text, textFormat, rect, dxBrush);
				}
				p1 = new Point(xCoordD, chartScale.GetYByValue(three82));
				p2 = new Point(xCoordD + oneToOneXAmt, chartScale.GetYByValue(three82));
				using (SharpDX.Direct2D1.Brush dxBrush = LineColorBrush.ToDxBrush(RenderTarget)) { 	
					RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush);
				}
				using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
					SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordD, chartScale.GetYByValue(six18), 200, 200);
					String text =  ".618 (Target)"; 							
					RenderTarget.DrawText(text, textFormat, rect, dxBrush);
				}
				p1 = new Point(xCoordD, chartScale.GetYByValue(six18));
				p2 = new Point(xCoordD + oneToOneXAmt, chartScale.GetYByValue(six18));
				using (SharpDX.Direct2D1.Brush dxBrush = LineColorBrush.ToDxBrush(RenderTarget)) { 	
					RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush);
				}
				p1 = new Point(xCoordD, chartScale.GetYByValue(priceD));
				p2 = new Point(xCoordD, chartScale.GetYByValue(priceA));
				using (SharpDX.Direct2D1.Brush dxBrush = LineColorBrush.ToDxBrush(RenderTarget)) { 	
					RenderTarget.DrawLine(p1.ToVector2(), p2.ToVector2(), dxBrush);
				}
				
				//Print((foundGartley ? "Gartley " : "") + (foundBat ? "Bat " : "") + (foundCrab ? "Crab " : "") + (foundButterfly ? "Buterfly " : ""));
				if(foundGartley || foundBat || foundCrab || foundButterfly){
					offset = bullish ? offset + 10 : offset -10;
					using (SharpDX.Direct2D1.Brush dxBrush = TextColorBrush.ToDxBrush(RenderTarget)) { 					
						SharpDX.RectangleF rect = new SharpDX.RectangleF((float)xCoordX-10, (float)yCoordX + offset, 200, 200);
						String text ="";
						if(foundBat && foundCrab) text = "Bat / Crab";
						else if(foundGartley && foundButterfly) text = "Gartley / Butterfly";
						else if(foundGartley && foundCrab) text = "Gartley / Crab";
						else if(foundBat) text = "Bat";
						else if (foundCrab) text = "Crab";
						else if (foundGartley) text = "Gartley";
						else if(foundButterfly) text = "Butterfly";
						
						RenderTarget.DrawText(text, textFormat, rect, dxBrush);
					}
				}
			}
			
			
			
		}
		
		protected override void OnStateChange()
		{
			if (State == State.SetDefaults)
			{
				StartAnchor		= new ChartAnchor { DisplayName = Custom.Resource.NinjaScriptDrawingToolAnchorStart,	IsEditing = true, DrawingTool = this };
				MiddleAnchor	= new ChartAnchor { DisplayName = Custom.Resource.NinjaScriptDrawingToolAnchorMiddle,	IsEditing = true, DrawingTool = this };
				EndAnchor		= new ChartAnchor { DisplayName = Custom.Resource.NinjaScriptDrawingToolAnchorEnd,		IsEditing = true, DrawingTool = this };
				MiddleAnchor2	= new ChartAnchor { DisplayName = Custom.Resource.NinjaScriptDrawingToolAnchorMiddle+"2",	IsEditing = true, DrawingTool = this };
				EndAnchor2		= new ChartAnchor { DisplayName = Custom.Resource.NinjaScriptDrawingToolAnchorEnd+"2",		IsEditing = true, DrawingTool = this };				
				DrawingState	= DrawingState.Building;
				AreaBrush		= Brushes.Gainsboro;
				AreaOpacity		= 10;
				OutlineStroke	= new Stroke(LineColorBrush, 1f);
				ShapeType		= ChartShapeType2.Unset;
				MiddleAnchor.IsBrowsable = false;
				MiddleAnchor2.IsBrowsable = false;
				TextColorBrush = Brushes.DarkOrange;
				LineColorBrush = Brushes.Gray;
				OneToOneBrush = Brushes.DimGray;
				AOIBrush = Brushes.RosyBrown;
				//SnapToHighLow = false;
				SCRules = false;
				FibTolerance = .03;
				ShowABOnetoOne = true;
				showGartley = true;
				showBat = true;
				showCrab = true;
				showButterfly = true;
			}
			else if (State == State.Terminated)
				Dispose();
		}
	}

	
	/// <summary>
	/// Represents an interface that exposes information regarding a Triangle IDrawingTool.
	/// </summary>
	[EditorBrowsable(EditorBrowsableState.Always)]
	public class Triangle2 : BmdXABCD
	{
		public override object Icon { get { return Gui.Tools.Icons.DrawTriangle; } }

		protected override void OnStateChange()
		{			
			base.OnStateChange();
			if (State == State.SetDefaults)
			{
				Name		= "bmdXABCD";
				ShapeType	= ChartShapeType2.Triangle2;
				MiddleAnchor.IsBrowsable = true;
				MiddleAnchor2.IsBrowsable = true;
			}
		}
	}

	public static partial class Draw
	{
		

		


	}
}
